/**
 *  @file       commands/run.js
 *  @author     Eddie Roosenmaallen <silvermoon82@gmail.com>
 *  @date       November 2022
 *
 *  Fetch and archive the authenticated user's stream
 */

'use strict';

const fs = require('fs');
const path = require('path');
const readline = require('readline');
const { MastoBot } = require('mastobot');

const {
  loadConfig,
  saveConfigFile,
  debugging,
} = require('../lib/util');


exports.command = 'run [directory]';

exports.describe = 'Download and save any toots older or newer than already saved';

exports.builder = {
  'directory': {
    describe: 'The directory in which to store the configuration and data',
    default: '.',
    type: 'path',
  },
  'verbose': {
    describe: 'Be more chatty about what\'s being done',
    default: false,
    type: 'boolean',
  },
  'quiet': {
    describe: 'Be less chatty about what\'s happening',
    default: false,
    type: 'boolean',
  },
  'refetch': {
    describe: 'If set, will ignore toots it has already downloaded, and will fetch the user\'s entire history',
  default: false,
    type: 'boolean',
  },
};


exports.handler = async function run$$handler(argv) {
  const config = loadConfig(argv);
  const directory = path.dirname(config._path);
  
  const { api_domain, api_token } = config;
  const api_url = new URL('https://' + config.api_domain).href;

  const mb = new MastoBot(api_url, { api_token });
  const me = (await mb.checkCredentials()).data;


  console.info(`Connected as @${me.acct}@${api_domain}`);

  let { newest_id, oldest_id, next, prev } = config;

  let stop = false;
  let downloaded = 0, saved = 0;

  let page;

  // If we have a "prev" link (from our newest Toot), follow that way first 
  if (config.prev)
  {
    argv.quiet || console.log('Checking for newer toots...');
    const { prev, next } = await fetchToots(await mb.get(config.prev), 'prev');
    config.prev = prev || config.prev;
    // config.next = config.next || next;
  }

  if (config.next && !argv.refetch)
  {
    argv.quiet || console.log('Looking for older toots...');
    const { next } = await fetchToots(await mb.get(config.next));
    config.next = next || config.next;
  }
  else
  {
    argv.quiet || console.log('Fetching all your toots...')
    const links = await fetchToots(await mb.getAccountStatuses(me.id));
    Object.assign(config, links);
  }

  const savedConfig = saveConfigFile(config, argv);
  console.log('Saved config:', savedConfig);
  
  argv.verbose && console.log(`Downloaded ${downloaded} toots in total, saved ${saved} new`);

  async function fetchToots(page, direction = 'next')
  {
    function stopFetching() {
      process.off('SIGINT', stopFetching);
      process.off('SIGQUIT', stopFetching);
      console.log('Stopping...');
      stop = true;
    }

    process.once('SIGINT', stopFetching);
    process.once('SIGQUIT', stopFetching);

    const links = {
      prev: config.prev || null,
      next: config.next || null,
    };

    do {
      const newest = page.data[0];
      const [oldest] = page.data.slice(-1);

      argv.quiet || console.log(`... got ${page.data.length} toots from`, oldest?.created_at.toLocaleString(), 'to', newest?.created_at.toLocaleString());

      const toots = page.data;
      downloaded += toots.length;

      let savedNew = 0;

      toots.forEach(toot => {
        const d = new Date(toot.created_at);
        const subdir = path.join('' + d.getFullYear(), (d.getMonth() + 1).toString(10).padStart(2, '0'));
        const tootDir = path.resolve(directory, subdir);
        const tootFileName = toot.id + '.json';
        const tootFile = path.resolve(tootDir, tootFileName);

        if (toot.account.id === me.id)
          toot.account = me.url;

        if (!config.newest_id || toot.id > config.newest_id)
        {
          config.newest_id = toot.id;
          links.prev = page.link?.prev || links.prev;
        }
        if (!config.oldest_id || toot.id < config.oldest_id)
        {
          config.oldest_id = toot.id;
          links.next = page.link?.next || links.next;
        }

        if (!fs.existsSync(tootDir))
          fs.mkdirSync(tootDir, { recursive: true });

        if (!fs.existsSync(tootFile))
        {
          fs.writeFileSync(tootFile, JSON.stringify(toot, null, 2));
          debugging('saves') && console.debug(`  - wrote ${path.join(subdir, tootFileName)}`);
          saved++;
          savedNew++;
        }
        else
        {
          debugging('saves') && console.debug(`  - skip ${path.join(subdir, tootFileName)}; exists`);
        }
      });

      if (savedNew && argv.verbose)
        console.info(`    saved ${savedNew} new toots`);

      // Throttle our requests some so we don't anger instance admins
      await new Promise(resolve => setTimeout(resolve, 500));
    }
    while (
      !stop &&
      page.data.length &&
      page.link[direction] &&
      (page = await mb.get(page.link[direction])) &&
      page.data.length
    );

    process.off('SIGINT', stopFetching);
    process.off('SIGQUIT', stopFetching);

    return links;
  }
};

