/**
 *  @file       lib/util.js
 *  @author     Eddie Roosenmaallen <silvermoon82@gmail.com>
 *  @date       November 2022
 *
 *  This module provides some uitility functions
 */

'use strict';

const fs = require('fs');
const path = require('path');

const exitCodes = {
  EXIT_OK:            0,  // no error
  EXIT_ERROR:         1,  // Error, not otherwise specified
  EXIT_ERROR_CONFIG:  2,  // Failed to load local config
};

const defaultConfig = {
  _path: null,
  api_domain: null,
  api_token: null,
};

/**
 *  Copy nonnull enumerable own properties from `sources` onto `dest`
 *
 *  @param  {Object}    dest    Destination object. Will be mutated.
 *  @param  {...Object} sources Source objects. Properties will be assigned
 *    left-first, so in the case of duplicate property names, the
 *    last-listed value will win
 *
 *  @return {Object}            dest, with new properties
 */
function objectAssignNonnull(dest, ...sources)
{
  sources.forEach(src => {
    for (let [p, v] of Object.entries(src))
    {
      if (typeof v === 'undefined' || v === null)
        continue;
      dest[p] = v;
    }
  });

  return dest;
}


/**
 *  Copy nonnull enumerable own properties from `sources` onto `dest`,
 *  without clobbering existing properties.
 *
 *  @param  {Object}    dest    Destination object. Will be mutated.
 *  @param  {...Object} sources Source objects. Properties will be assigned
 *    left-first, so in the case of duplicate property names, the
 *    first-listed value will win
 *
 *  @return {Object}            dest, with new properties
 */
function objectAssignNoclobber(dest, ...sources)
{
  sources.forEach(src => {
    for (let [p, v] of Object.entries(src))
    {
      if (typeof v === 'undefined' || v === null)
        continue;
      if (typeof dest[p] !== 'undefined')
        continue;
      dest[p] = v;
    }
  });

  return dest;
}


/**
 *  Handle a fatal error processing a command
 *
 *  @param  {Error} error Error caught by cli framework
 */
function handleCommandError(msg, error, yargs)
{
  const code = '' + (error.code || null);
  console.log({msg, error});

  if (code.startsWith('ESYNTAX.'))
  {
    console.error(`Fatal error: ${error.message}
Filename: ${error.fileName}
Position: ${error.column}`);
    process.exit(EXIT_ERROR_CONFIG);
  }

  console.error(`Unexpected ${error.name}:`, error.message);
  console.error(error.stack);
  process.exit(EXIT_ERROR);
}


/**
 *  Load a specified configuration file
 *
 *  @param  {String} filePath Path to the config file to open
 *  @return {Object}          Configuration parsed from the file
 */
function loadConfigFile(filePath = './masto-dl.conf')
{
  filePath = path.resolve(filePath);
  const config = Object.assign({}, defaultConfig);

  // If local config doesn't exist, return empty object
  if (!fs.existsSync(filePath))
    return config;

  try
  {
    const fileConfig = JSON.parse(fs.readFileSync(filePath, 'utf8'));
    objectAssignNonnull(config, fileConfig);
    config._path = filePath;
  }
  catch (error)
  {
    const msg = '' + error.message;

    error.message = 'Error loading configuration: ' + error.message;
    error.fileName = filePath;

    if (error instanceof SyntaxError)
    {
      const [, column] = msg.match(/on line ([0-9]+)$/);

      if (column)
        error.column = column;
    }
    // else if (other predictable error) {} ...
    else
    {
      console.error(`Unexpected ${error.name} when loading configuration: ${error.message}`, '\n' + error.stack);
    }
  }

  return config;
}


/**
 *  Load a local configuration, given a parsed options object
 *
 *  @param  {Object|String} argv  Parsed CLI options from yargs, or string path
 *  @return {Object}              Configuration from loadConfigFile
 */
function loadConfig(argv)
{
  const configFile = getConfigPath(argv);
  const config = loadConfigFile(configFile);



  config._path = configFile;

  return config;
}

/**
 *  Determine path to config file, given an argv object containing directory
 *    and configFile properties
 *
 *  @param  {Object|String} argv             Arguments parsed from
 *                                           commandline, or string path
 *  @param  {Object?}       argv.directory   Working directory
 *  @param  {Object?}       argv.configFile  Name of config file
 *  @return {String}                         Absolute path to the config file
 */
function getConfigPath(argv)
{
  const basePath = (typeof argv === 'string')
    ? path.resolve(argv)
    : path.resolve(argv.directory);

  if (!fs.existsSync(basePath))
  {
    console.error(`Error: Configuration path "${basePath}" does not exist`);
    process.exit(exitCodes.EXIT_CONFIG_NOTDIR);
  }

  return path.resolve(basePath, argv.configFile);
}

/**
 *  Save a config file at a specified path
 *
 *  @param  {Object}        config    Configuration options (will not be mutated in place)
 *  @param  {String|Object} filePath  Path to save to, or argv object to determine it from
 *  @return {Object}                  config object as written, with path property updated
 */
function saveConfigFile(config, filePath)
{
  if (!filePath)
    filePath = getConfigPath(config._path || './masto-dl.conf');
  else
    filePath = getConfigPath(filePath);

  const configToWrite = objectAssignNonnull(Object.assign({}, defaultConfig), config);
  delete configToWrite._path,

  fs.writeFileSync(filePath, JSON.stringify(configToWrite, null, 2));

  configToWrite._path = filePath;

  return configToWrite;
}


/**
 *  Check whether $DEBUG env var is active and matches an optional tag
 *
 *  @param  {string} tag Optional tag to match against tokens in $DEBUG
 *  @return {boolean}     True if debugging is active and tag is unspecified
 *    or is present in environment
 */
function debugging(tag)
{
  // if $DEBUG isn't set, bail early with 'no'
  if (!process.env.DEBUG)
    return false;

  // if tag isn't set, bail early with 'yes'
  if (!tag)
    return true;

  const tags = process.env.DEBUG.split(',').map(t => t.toLowerCase());

  if (tags.includes(tag.toLowerCase()))
    return true;

  return false;
}


Object.assign(exports, {
  objectAssignNonnull,
  objectAssignNoclobber,
  loadConfig,
  loadConfigFile,
  saveConfigFile,
  getConfigPath,
  debugging,
  handleCommandError,
});
